/*
	CASM.C	-- WRITTEN BY LEOR ZOLMAN, 2/82
	MODIFIED FOR V1.50 11/14/82

	CP/M ASM PREPROCESSOR: RENDERS MAC.COM AND CMAC.LIB UNNECESSARY.

	SEE THE CASM APPENDIX IN THE USER'S GUIDE FOR COMPLETE INFORMATION.

	COMPILE AND LINK WITH:

		CC CASM.C -O -E5000

		L2 CASM
	(OR)	CLINK CASM
*/

#INCLUDE <STDIO.H>

#DEFINE TITLE "BD SOFTWARE CRL-FORMAT ASM PREPROCESSOR V1.50\N"


/*
 *	CUSTOMIZABLE DEFINITIONS:
 */

#DEFINE DEFUSER	""		/* DEFAULT USER AREA FOR INCLUDE FILES	*/
				/* MAKE THIS A NULL STRING FOR "CURRENT" */

#DEFINE DEFDISK ""		/* DEFAULT DISK FOR INCLUDE FILES	*/
#DEFINE CASMEXT	".CSM"		/* EXTENSION ON INPUT FILES 	*/
#DEFINE ASMEXT	".ASM"		/* EXTENSION ON OUTPUT FILES	*/
#DEFINE SUBFILE "A:$$$.SUB"	/* SUBMIT FILE TO ERASE ON ERROR. TO NOT */
				/* ERASE ANY, USE A NULL STRING ("")	 */

#DEFINE CONTROL_C 3		/* ABORT CHARACTER */
#DEFINE	EQUMAX	500		/* MAXIMUM NUMBER OF EQU OPS	*/
#DEFINE FUNCMAX	100		/* MAXIMUM NUMBER OF FUNCTIONS  */
#DEFINE NFMAX	100		/* MAXIMUM NUMBER OF EXTERNAL
				   FUNCTIONS IN ONE FUNCTION 	*/
#DEFINE LABMAX	150		/* MAX NUMBER OF LOCAL LABELS IN ONE FUNC */
#DEFINE TXTBUFSIZE 2000		/* MAX # OF CHARS FOR LABELS AND NEEDED
				   FUNCTION NAMES FOR A SINGLE FUNCTION	*/

/*
 *	END OF CUSTOMIZABLE SECTION
 */

#DEFINE DIRSIZE	512		/* MAX # OF BYTE IN CRL DIRECTORY 	*/
#DEFINE TPALOC	0X100		/* BASE OF TPA IN YOUR SYSTEM 	*/

		/* GLOBAL DATA USED THROUGHOUT PROCESSING
		   OF THE INPUT FILE:			*/

CHAR	FBUF[BUFSIZ];		/* I/O BUFFER FOR MAIN INPUT FILE	*/
CHAR	INCBUF[BUFSIZ];		/* I/O BUFFER FOR INCLUDED FILE 	*/
CHAR	OBUF[BUFSIZ];		/* I/O BUFFER FOR OUTPUT FILE		*/

CHAR	*CBUFP;			/* POINTER TO CURRENTLY ACTIVE INPUT BUF */
CHAR	*CFILNAM;		/* POINTER TO NAME OF CURRENT INPUT FILE */
CHAR	NAMBUF[30],		/* FILENAMES FOR CURRENT INTPUT */
	NAMBUF2[30],		/* AND OUTPUT FILES.		*/
	ONAMBUF[30];

CHAR	*EQUTAB[EQUMAX];	/* TABLE OF ABSOLUTE SYMBOLS	*/
INT	EQUCOUNT;		/* # OF ENTRIES IN EQUTAB	*/

CHAR	*FNAMES[FUNCMAX];	/* LIST OF FUNCTIONS IN THE SOURCE FILE */
INT	FCOUNT;			/* # OF ENTRIES IN FNAMES		*/

INT	LINO,SAVLINO;		/* LINE NUMBER VALUES USED FOR ERROR 	*/
				/* REPORTING.				*/

CHAR	DOINGFUNC;		/* TRUE IF CURRENTLY PROCESSING A FUNCTION */

CHAR	ERRF;			/* TRUE IF AN ERROR HAS BEEN DETECTED	*/
CHAR	VERBOSE;		/* TRUE TO INSERT WORDY COMMENTS IN OUTPUT */
CHAR	CAREFUL;		/* TRUE TO DETECT OLD CMAC.LIB MACROS   */
CHAR	BLANKFLAG;		/* TRUE IF LAST LINE PROCESSED WAS NULL */

		/* GLOBAL DATA USED DURING THE PROCESSING OF A
		   SINGLE FUNCTION IN THE SOURCE FILE:		*/

CHAR	*NFLIST[NFMAX];		/* LIST OF NEEDED FUNCTIONS FOR A FUNCTION */
INT	NFCOUNT;		/* NUMBER OF ENTRIES IN NFLIST	*/

STRUCT {
	CHAR *LABNAM;		/* NAME OF FUNCTION LABEL */
	CHAR DEFINED;		/* WHETHER IT HAS BEEN DEFINED YET */
} LABLIST[LABMAX];

INT	LABCOUNT;		/* NUMBER OF LOCAL LABELS IN A FUNCTION */

CHAR	TXTBUF[TXTBUFSIZE],	/* WHERE TEXT OF NEEDED FUNCTION NAMES	*/
	*TXTBUFP;		/* AND FUNCTION LABELS GO		*/

CHAR 	LINBUF[150],		/* TEXT LINE BUFFERS	*/
	LINSAV[150],
	WORKBUF[150],
	PBUF[150], *PBUFP;

CHAR	*CFUNAM;		/* POINTER TO NAME OF CURRENT FUNCTION */
INT	RELBLC;			/* RELOCATION OBJECT COUNT FOR A FUNCTION */

CHAR	PASTNFS;		/* TRUE IF WE'VE PASSED ALL NEEDED FUNCTION */
				/* DECLARATIONS ("EXTERNAL" PSEUDO OPS)	    */

INT	ARGCNT;			/* VALUES SET BY THE "PARSE_LINE" FUNCTION */
CHAR	*LABEL,
	*OP,
	*ARGSP,
	*ARGS[40];

CHAR 	*GPCPTR;		/* GENERAL-PURPOSE TEXT POINTER	*/

/*
 * OPEN MAIN INPUT FILE, OPEN OUTPUT FILE, INITIALIZE NEEDED GLOBALS
 * AND PROCESS THE FILE:
 */

MAIN(AARGHC,AARGHV)
CHAR **AARGHV;
{
	INT I,J,K;
	CHAR C, *INPNAM, *OUTNAM;

	PUTS(TITLE);

	INITEQU();		/* INITIALIZE EQU TABLE WITH RESERVED WORDS */
	FCOUNT = 0;		/* HAVEN'T SEEN ANY FUNCTIONS YET */
	DOINGFUNC = FALSE;	/* NOT CURRENTLY PROCESSING A FUNCTION */
	ERRF = FALSE;		/* NO ERRORS YET */
	VERBOSE = CAREFUL = FALSE;
	INPNAM = OUTNAM = NULL;		/* HAVEN'T SEEN ANY NAMES YET */
	BLANKFLAG = FALSE;	/* HAVEN'T JUST PROCESSED A NULL LINE */
	
	WHILE (--AARGHC) 
	{
		++AARGHV;		/* BUMP TO NEXT ARG TEXT */
		IF (**AARGHV == '-')
		{
		    SWITCH(C = AARGHV[0][1])
		    {
			CASE 'F':
				CAREFUL = 1;
				BREAK;

			CASE 'C':
				VERBOSE = 1;
				BREAK;

			CASE 'O':
				IF (AARGHV[0][2])
					OUTNAM = &AARGHV[0][2];
				ELSE IF (--AARGHC)
					OUTNAM = *++AARGHV;
				ELSE GOTO USAGE;
				BREAK;

			DEFAULT: GOTO USAGE;
		    }
		}
		ELSE
			INPNAM = *AARGHV;
	}

	IF (!INPNAM) {
  USAGE:	PUTS("USAGE:\TCASM [-F] [-C] [-O <NAME>] <FILENAME>\N");
		PUTS("-F: FLAG OLD CMAC.LIB MACROS IF SPOTTED\N");
		PUTS("-C: DON'T STRIP COMMENTS FROM INPUT AND OUTPUT\N");
		PUTS("-O <NAME>: CALL THE OUTPUT FILE <NAME>.ASM\N");
		EXIT();
	}

				/* SET UP FILENAMES WITH PROPER EXTENSIONS: */
	FOR (I = 0; (C = INPNAM[I]) && C != '.'; I++)
		NAMBUF[I] = C;
	NAMBUF[I] = '\0';

	STRCPY(ONAMBUF, OUTNAM ? OUTNAM : NAMBUF);
	STRCAT(NAMBUF,CASMEXT);		/* INPUT FILENAME */
	CBUFP = FBUF;			/* BUFFER POINTER */
	CFILNAM = NAMBUF;		/* CURRENT FILENAME POINTER */
	IF (FOPEN(CFILNAM,CBUFP) == ERROR)
		EXIT(PRINTF("CAN'T OPEN %S\N",CFILNAM));

	IF (!HASDOT(ONAMBUF))
		STRCAT(ONAMBUF,ASMEXT);		/* OUTPUT FILENAME */
	IF (FCREAT(ONAMBUF,OBUF) == ERROR)
		EXIT(PRINTF("CAN'T CREATE %S\N",ONAMBUF));

					/* BEGIN WRITING OUTPUT FILE */
	FPRINTF2(OBUF,"\NTPALOC\T\TEQU\T%04XH\N",TPALOC);

	LINO = 1;			/* INITIALIZE LINE COUNT */

	WHILE (GET_LINE()) {		/* MAIN LOOP */
		IF (KBHIT() && GETCHAR() == CONTROL_C)
			ABORT("ABORTED BY ^C\N");
		PROCESS_LINE();		/* PROCESS LINES TILL EOF */
		LINO++;
	}

	IF (DOINGFUNC)			/* IF ENDS INSIDE A FUNCTION, ERROR */
		ABORT("FILE ENDS, BUT LAST FUNCTION IS UNTERMINATED\N");

	IF (ERRF)
	{
		PUTS("FIX THOSE ERRORS AND TRY AGAIN...");
		UNLINK(ONAMBUF);
		IF (*SUBFILE) 
			UNLINK(SUBFILE);
	}
	ELSE
	{
							/* END OF FUNCTIONS */
		FPUTS2("\NEND$CRL\T\TEQU\T$-TPALOC\N",OBUF);
		FPUTS2("SECTORS$ EQU ($-TPALOC)/256+1 ;USE FOR \"SAVE\" !.\N",
			OBUF);
		PUTDIR();			/* NOW PUT OUT CRL DIRECTORY */
		FPUTS2("\T\TEND\N",OBUF);	/* END OF ASM FILE */
		PUTC(CPMEOF,OBUF);		/* CP/M EOF CHARACTER */
		FCLOSE(CBUFP);			/* CLOSE INPUT FILE */
		FCLOSE(OBUF);			/* CLOSE OUTPUT FILE */
		PRINTF("%S IS READY TO BE ASSEMBLED.\N",ONAMBUF);
	}
}

/*
 * GET A LINE OF TEXT FROM INPUT STREAM, AND PROCESS
 * "INCLUDE" OPS ON THE FLY:
 */

INT GET_LINE()
{
	INT I;

TOP:	IF (!FGETS(LINBUF,CBUFP)) {		/* ON EOF: */
		IF (CBUFP == INCBUF) {		/* IN AN "INCLUDE" FILE? */
			FABORT(CBUFP->_FD);		/* CLOSE THE FILE */
			CBUFP = FBUF;		/* GO BACK TO MAINLINE FILE */
			CFILNAM = NAMBUF;
			LINO = SAVLINO + 1;
			RETURN GET_LINE();
		}
		ELSE RETURN NULL;
	}

	IF (!VERBOSE)				/* STRIP COMMMENTS, DEFAULT */
	{
		FOR (I = 0; LINBUF[I]; I++)
		{
			IF (LINBUF[I] == ';')
			{
				WHILE (I && ISSPACE(LINBUF[I-1]))
					I--;
				IF (!I && BLANKFLAG)
				{
					LINO++;
					GOTO TOP;
				}
				STRCPY(&LINBUF[I], "\N");
				BLANKFLAG = TRUE;
				BREAK;
			}
			IF (LINBUF[I] == '\'' || LINBUF[I] == '"')
				BREAK;
		}
		IF (!LINBUF[I])
			BLANKFLAG = FALSE;		
	}

	PARSE_LINE();				/* NOT EOF. PARSE LINE */
	IF (STREQ(OP,"INCLUDE")  ||		/* CHECK FOR FILE INCLUSION */
	    STREQ(OP,"MACLIB")) {
		IF (CBUFP == INCBUF)		/* IF ALREADY IN AN INCLUDE, */
		 ABORT("ONLY ONE LEVEL OF INCLUSION IS SUPPORTED"); /* ERROR */
		IF (!ARGSP)
		 ABORT("NO FILENAME SPECIFIED");
		CBUFP = INCBUF;			/* SET UP FOR INCLUSION */
		SAVLINO = LINO;

		FOR (I = 0; !ISSPACE(ARGSP[I]); I++)	/* PUT NULL AFTER */
			;				/* FILENAME	  */
		ARGSP[I] = '\0';

		*NAMBUF2 = '\0';

		IF (*ARGSP == '<') {		/* LOOK FOR MAGIC DELIMITERS */
			STRCPY(NAMBUF2,DEFUSER);
			IF (ARGSP[2] != ':')	/* IF NO EXPLICIT DISK GIVEN */
				STRCAT(NAMBUF2,DEFDISK); /* THEN USE DEFAULT */
			STRCAT(NAMBUF2,ARGSP+1);
			IF (NAMBUF2[I = STRLEN(NAMBUF2) - 1] == '>')
				NAMBUF2[I] = '\0';
		} ELSE IF (*ARGSP == '"') {
			STRCPY(NAMBUF2,ARGSP+1);
			IF (NAMBUF2[I = STRLEN(NAMBUF2) - 1] == '"')
				NAMBUF2[I] = '\0';
		} ELSE
			STRCPY(NAMBUF2,ARGSP);

		IF (FOPEN(NAMBUF2,CBUFP) == ERROR) {
			IF (!HASDOT(NAMBUF2)) {
				STRCAT(NAMBUF2,".LIB");
				IF (FOPEN(NAMBUF2,CBUFP) != ERROR)
					GOTO OK;
			}			    
			PRINTF("CAN'T OPEN %S\N",NAMBUF2);
			ABORT("MISSING INCLUDE FILE");
		}

	OK:	LINO = 1;
		CFILNAM = NAMBUF2;
		RETURN GET_LINE();
	}
	RETURN 1;
}

PARSE_LINE()
{
	INT I;
	CHAR C;

	LABEL = OP = ARGSP = NULL;
	ARGCNT = 0;

	STRCPY2(PBUF,LINBUF);
	STRCPY2(LINSAV,LINBUF);
	PBUFP = PBUF;

	IF (!*PBUFP) RETURN;		/* IGNORE NULL LINES */
	IF (!ISSPACE(C = *PBUFP)) {
		IF (C == ';')
			RETURN;		/* TOTALLY IGNORE COMMENT LINES */
		LABEL = PBUFP;		/* SET POINTER TO LABEL	*/
		WHILE (ISIDCHR(*PBUFP))	/* PASS OVER THE LABEL IDENTIFIER */
			PBUFP++;
		*PBUFP++ = '\0';	/* PLACE NULL AFTER THE IDENTIFIER */
	}

	SKIP_WSP(&PBUFP);
	IF (!*PBUFP || *PBUFP == ';')
		RETURN;
	OP = PBUFP;			/* SET POINTER TO OPERATION MNEMONIC */
	WHILE (ISALPHA(*PBUFP))
		PBUFP++;  		/* SKIP OVER THE OP 		*/
	IF (*PBUFP) *PBUFP++ = '\0';	/* PLACE NULL AFTER THE OP	*/


					/* NOW PROCESS ARGUMENTS	*/
	SKIP_WSP(&PBUFP);
	IF (!*PBUFP || *PBUFP == ';')
		RETURN;
	ARGSP = LINSAV + (PBUFP - PBUF);	/* SET POINTER TO ARG LIST */

					/* CREATE VECTOR OF PTRS TO ALL ARGS
					   THAT ARE POSSIBLY RELOCATABLE */
	FOR (ARGCNT = 0; ARGCNT < 40;) {
		WHILE (!ISIDSTRT(C = *PBUFP))
			IF (!C || C == ';')
				RETURN;
			ELSE
				PBUFP++;

		IF (ISIDCHR(*(PBUFP - 1))) {
			PBUFP++;
			CONTINUE;
		}

		ARGS[ARGCNT++] = PBUFP;			
		WHILE (ISIDCHR(*PBUFP)) PBUFP++;
		IF (*PBUFP) *PBUFP++ = '\0';
	}
	ERROR("TOO MANY OPERANDS IN THIS INSTRUCTION FOR ME TO HANDLE\N");
}

PROCESS_LINE()
{
	CHAR *CPTR, C;
	INT I,J;

	IF (OP) {
			/* CHECK FOR DEFINITIONS OF GLOBAL DATA THAT WILL BE
			   EXEMPT FROM RELOCATION WHEN ENCOUNTERED IN THE
			   ARGUMENT FIELD OF ASSEMBLY INSTRUCTIONS: 	   */

	   IF (STREQ(OP,"EQU") || STREQ(OP,"SET") ||
		(!DOINGFUNC &&
			(STREQ(OP,"DS") || STREQ(OP,"DB") || STREQ(OP,"DW"))))
	   {
		FPUTS2(LINBUF,OBUF);
		CPTR = SBRK2(STRLEN(LABEL) + 1);
		STRCPY(CPTR,LABEL);
		EQUTAB[EQUCOUNT++] = CPTR;
		IF (EQUCOUNT >= EQUMAX)
			ABORT(
		  "TOO MANY EQU LINES...INCREASE 'EQUMAX' AND RECOMPILE CASM");
		RETURN;
	   }

	   IF (STREQ(OP,"EXTERNAL")) {
		IF (!DOINGFUNC) ABORT(
		 "'EXTERNAL'S FOR A FUNCTION MUST APPEAR INSIDE THE FUNCTION");
		IF (PASTNFS) ERROR(
		 "EXTERNALS MUST ALL BE TOGETHER AT START OF FUNCTION\N");
		FOR (I = 0; I < ARGCNT; I++) {
			NFLIST[NFCOUNT++] = TXTBUFP;
			STRCPY(TXTBUFP,ARGS[I]);
			BUMPTXTP(ARGS[I]);
		}
		IF (NFCOUNT >= NFMAX) {
		  PRINTF("TOO MANY EXTERNAL FUNCTIONS IN FUNCTION \"%S\"\N",
					CFUNAM);
		  ABORT("CHANGE THE NFMAX CONSTANT AND RECOMPILE CASM");
		}
		RETURN;
	   }

	   IF (STREQ(OP,"FUNCTION")) {
		IF (!FCOUNT) {
			IF (VERBOSE)
			 FPUTS2("\N; DUMMY EXTERNAL DATA INFORMATION:\N",OBUF);

			FPUTS2("\T\TORG\TTPALOC+200H\N",OBUF);
			FPUTS2("\T\TDB\T0,0,0,0,0\N",OBUF);
		}

		IF (DOINGFUNC) {
			PRINTF("'FUNCTION' OP ENCOUNTERED IN A FUNCTION.\N");
			ABORT("DID YOU FORGET AN 'ENDFUNC' OP?");
		}
		IF (!ARGCNT)
			ABORT("A NAME IS REQUIRED FOR THE 'FUNCTION' OP");

		CFUNAM = SBRK2(STRLEN(ARGS[0]) + 1);
		FNAMES[FCOUNT++] = CFUNAM;
		STRCPY(CFUNAM,ARGS[0]);

		PRINTF("PROCESSING THE %S FUNCTION...          \R",CFUNAM);

		DOINGFUNC = 1;
		TXTBUFP = TXTBUF;
		LABCOUNT = 0;
		NFCOUNT = 0;
		PASTNFS = 0;

		IF (VERBOSE)
			FPRINTF2(OBUF,"\N\N; THE \"%S\" FUNCTION:\N",CFUNAM);

		FPRINTF2(OBUF,"%S$BEG\TEQU\T$-TPALOC\N",CFUNAM);
		RETURN;
	   }

	   IF (STREQ(OP,"ENDFUNC") || STREQ(OP,"ENDFUNCTION")) {
		IF (!DOINGFUNC)
		  ABORT("'ENDFUNC' OP ENCOUNTERED WHILE NOT IN A FUNCTION");

		IF (!PASTNFS) FLUSHNFS();    /* FLUSH NEEDED FUNCTION LIST */
		FPRINTF2(OBUF,"%S$END\TEQU\T$\N",CFUNAM);
		DORELOC();		     /* FLUSH RELOCATION PARAMETERS */

		FOR (I = 0; I < LABCOUNT; I++)	/* DETECT UNDEFINED LABELS */
		  IF (!LABLIST[I].DEFINED) {
			PRINTF("THE LABEL %S IN FUNCTION %S IS UNDEFINED\N",
					LABLIST[I].LABNAM,CFUNAM);
			ERRF = 1;
		  }
		DOINGFUNC = 0;
		RETURN;
	   }
	}

	IF (CAREFUL)
	IF (STREQ(OP,"RELOC") || STREQ(OP,"DWREL") || STREQ(OP,"DIRECT") ||
	    STREQ(OP,"ENDDIR") || STREQ(OP,"EXREL") || STREQ(OP,"EXDWREL") ||
	    STREQ(OP,"PRELUDE") || STREQ(OP,"POSTLUDE") || STREQ(OP,"DEFINE"))
		ERROR("OLD MACRO '%S' LEFTOVER FROM \"CMAC.LIB\" DAYS...\N",
							OP);

				/* NO SPECIAL PSEUDO OPS, SO NOW PROCESS
				   THE LINE AS A LINE OF ASSEMBY CODE: 	*/

	IF (STREQ(OP,"END")) RETURN;		/* DON'T ALLOW "END" YET     */

	IF (!DOINGFUNC || (!LABEL && !OP))	/* IF NOTHING INTERESTING ON */
		RETURN FPUTS2(LINBUF,OBUF);	/* LINE, IGNORE IT	*/

	IF (!PASTNFS)				/* IF HAVEN'T FLUSHED NEEDED */
		FLUSHNFS();			/* FUNCTION LIST YET, DO IT  */

						/* CHECK FOR POSSIBLE LABEL  */
	IF (LABEL) {
		FPRINTF2(OBUF,"%S$L$%S\T\TEQU\T$-%S$STRT\N",
			CFUNAM, LABEL, CFUNAM);
		FOR (I=0; LINBUF[I]; I++)
			IF (ISSPACE(LINBUF[I]) || LINBUF[I] == ':')
				BREAK;
			ELSE
				LINBUF[I] = ' ';
		IF (LINBUF[I] == ':') LINBUF[I] = ' ';
		FOR (I = 0; I < LABCOUNT; I++)	  /* CHECK IF IN LABEL TABLE */
		  IF (STREQ(LABEL,LABLIST[I].LABNAM)) {	        /* IF FOUND, */
			IF (LABLIST[I].DEFINED) {  /* CHECK FOR REDEFINITION */
				ERROR("RE-DEFINED LABEL:");
				PRINTF("%S, IN FUNCTION %S\N",
						LABLIST[I].LABNAM,CFUNAM);
			}
			 ELSE
				LABLIST[I].DEFINED = 1;
			GOTO OUT;
		  }
		LABLIST[I].LABNAM = TXTBUFP;	/* ADD NEW ENTRY TO */
		LABLIST[I].DEFINED = 1;		/* LABEL LIST 	    */
		STRCPY(TXTBUFP,LABEL);
		BUMPTXTP(LABEL);
		LABCOUNT++;
	}
OUT:
	IF (!OP) RETURN FPUTS2(LINBUF,OBUF);	/* IF LABEL ONLY, ALL DONE   */

						/* IF A NON-RELOCATABLE OP,  */
	IF (NORELOP(OP)) RETURN FPUTS2(LINBUF,OBUF);	/* THEN WE'RE DONE   */

	IF (ARGCNT && DOINGFUNC)
	  FOR (I = 0; I < ARGCNT; I++) {
		IF (GPCPTR = ISEF(ARGS[I]))
		   SPRINTF(WORKBUF,"%S$EF$%S-%S$STRT",
				CFUNAM,GPCPTR,CFUNAM);
		ELSE IF (NOREL(ARGS[I])) CONTINUE;
		ELSE {
			SPRINTF(WORKBUF,"%S$L$%S",CFUNAM,ARGS[I]);
			FOR (J = 0; J < LABCOUNT; J++)
				IF (STREQ(ARGS[I],LABLIST[J].LABNAM))
					GOTO OUT2;
			LABLIST[J].LABNAM = TXTBUFP;	/* ADD NEW ENTRY TO */
			LABLIST[J].DEFINED = 0;		/* LABEL LIST 	    */
			STRCPY(TXTBUFP,ARGS[I]);
			BUMPTXTP(TXTBUFP);
			LABCOUNT++;
		}		   

	OUT2:
		REPLSTR(LINBUF, WORKBUF, ARGS[I] - PBUF, STRLEN(ARGS[I]));

		IF (STREQ(OP,"DW")) {
			FPRINTF2(OBUF,"%S$R%03D\TEQU\T$-%S$STRT\N",
				CFUNAM, RELBLC++, CFUNAM);
			IF (ARGCNT > 1)
			  ERROR("ONLY ONE RELOCATABLE VALUE ALLOWED PER DW\N");
		}
		ELSE
			FPRINTF2(OBUF,"%S$R%03D\TEQU\T$+1-%S$STRT\N",
				CFUNAM, RELBLC++, CFUNAM);
		BREAK;
	  }
	FPUTS2(LINBUF,OBUF);
}


/*
	TEST FOR OPS IN WHICH THERE IS GUANRANTEED TO BE NO NEED
	FOR GENERATION OF RELOCATION PARAMETERS. NOTE THAT THE LIST
	OF NON-RELOCATABLE OPS DOESN'T NECESSARILY HAVE TO BE COMPLETE,
	BECAUSE FOR ANY OP THAT DOESN'T MATCH, AN ARGUMENT MUST STILL
	PASS OTHER TESTS BEFORE IT IS DEEMED RELOCATABLE. THIS ONLY
	SPEEDS THINGS UP BY TELLING THE PROGRAM NOT TO BOTHER CHECKING
	THE ARGUMENTS.
*/

NORELOP(OP)
CHAR *OP;
{
	IF (STREQ(OP,"MOV")) RETURN 1;
	IF (STREQ(OP,"INR")) RETURN 1;
	IF (STREQ(OP,"DCR")) RETURN 1;
	IF (STREQ(OP,"INX")) RETURN 1;
	IF (STREQ(OP,"DCX")) RETURN 1;
	IF (STREQ(OP,"DAD")) RETURN 1;
	IF (STREQ(OP,"MVI")) RETURN 1;
	IF (STREQ(OP,"DB")) RETURN 1;
	IF (STREQ(OP,"DS")) RETURN 1;
	IF (OP[2] == 'I') {
		IF (STREQ(OP,"CPI")) RETURN 1;
		IF (STREQ(OP,"ORI")) RETURN 1;
		IF (STREQ(OP,"ANI")) RETURN 1;
		IF (STREQ(OP,"ADI")) RETURN 1;
		IF (STREQ(OP,"SUI")) RETURN 1;
		IF (STREQ(OP,"SBI")) RETURN 1;
		IF (STREQ(OP,"XRI")) RETURN 1;
		IF (STREQ(OP,"ACI")) RETURN 1;
	}
	IF (STREQ(OP,"ORG")) RETURN 1;
	IF (STREQ(OP,"TITLE")) RETURN 1;
	IF (STREQ(OP,"PAGE")) RETURN 1;
	IF (STREQ(OP,"IF")) RETURN 1;
	IF (STREQ(OP,"EJECT")) RETURN 1;
	IF (STREQ(OP,"MACRO")) RETURN 1;
	RETURN 0;
}


FLUSHNFS()
{
	INT I,J, LENGTH;

	PASTNFS = 1;
	RELBLC = 0;

	IF (VERBOSE)
		FPUTS2("\N\N; LIST OF NEEDED FUNCTIONS:\N",OBUF);

	FOR (I=0; I < NFCOUNT; I++) {
		STRCPY(WORKBUF,"\T\TDB\T'");
		LENGTH = STRLEN(NFLIST[I]);
		LENGTH = LENGTH < 8 ? LENGTH : 8;
		FOR (J = 0; J < LENGTH - 1; J++)
			WORKBUF[6+J] = NFLIST[I][J];
		WORKBUF[6+J] = '\0';
		FPRINTF2(OBUF,"%S','%C'+80H\N",WORKBUF,NFLIST[I][J]);
	}

	FPUTS2("\T\TDB\T0\N",OBUF);

	IF (VERBOSE)
		FPUTS2("\N; LENGTH OF BODY:\N",OBUF);

	FPRINTF2(OBUF,"\T\TDW\T%S$END-$-2\N",CFUNAM);

	IF (VERBOSE)
		FPUTS2("\N; BODY:\N",OBUF);

	FPRINTF2(OBUF,"%S$STRT\TEQU\T$\N",CFUNAM);
	IF (NFCOUNT) {
		FPRINTF2(OBUF,"%S$R%03D\TEQU\T$+1-%S$STRT\N",
			CFUNAM,RELBLC++,CFUNAM);
		FPRINTF2(OBUF,"\T\TJMP\T%S$STRTC-%S$STRT\N",CFUNAM,CFUNAM);
	}
	FPRINTF2(OBUF,"%S$EF$%S\TEQU\T%S$STRT\N",CFUNAM,CFUNAM,CFUNAM);
	FOR (I=0; I < NFCOUNT; I++)
		FPRINTF2(OBUF,"%S$EF$%S\TJMP\T0\N",CFUNAM,NFLIST[I]);
	FPRINTF2(OBUF,"\N%S$STRTC\TEQU\T$\N",CFUNAM);
}


DORELOC()
{
	INT I;

	IF(VERBOSE)
		FPUTS2("\N; RELOCATION PARAMETERS:\N",OBUF);

	FPRINTF2(OBUF,"\T\TDW\T%D\N",RELBLC);
	FOR(I = 0; I < RELBLC; I++)
		FPRINTF2(OBUF,"\T\TDW\T%S$R%03D\N",CFUNAM,I);
	FPUTS2("\N",OBUF);
}


PUTDIR()
{
	INT I,J, LENGTH;
	INT BYTECOUNT;

	BYTECOUNT = 0;

	FPUTS2("\N\T\TORG\TTPALOC\N\N; DIRECTORY:\N",OBUF);
	FOR (I = 0; I < FCOUNT; I++) {
		STRCPY(WORKBUF,"\T\TDB\T'");
		LENGTH = STRLEN(FNAMES[I]);
		LENGTH = LENGTH < 8 ? LENGTH : 8;
		FOR (J = 0; J < LENGTH - 1; J++)
			WORKBUF[6+J] = FNAMES[I][J];
		WORKBUF[6+J] = '\0';
		FPRINTF2(OBUF,"%S','%C'+80H\N",WORKBUF,FNAMES[I][J]);
		FPRINTF2(OBUF,"\T\TDW\T%S$BEG\N",FNAMES[I]);
		BYTECOUNT += (LENGTH + 2);
	}
	FPUTS2("\T\TDB\T80H\N\T\TDW\TEND$CRL\N",OBUF);

	BYTECOUNT += 3;
	IF (BYTECOUNT > DIRSIZE) {
		PRINTF("CRL DIRECTORY SIZE WILL EXCEED 512 BYTES;\N");
		PRINTF("BREAK THE FILE UP INTO SMALLER CHUNKS, PLEASE!\N");
		EXIT(-1);
	}
}


INITEQU()
{
	EQUTAB[0] = "A";
	EQUTAB[1] = "B";
	EQUTAB[2] = "C";
	EQUTAB[3] = "D";
	EQUTAB[4] = "E";
	EQUTAB[5] = "H";
	EQUTAB[6] = "L";
	EQUTAB[7] = "M";
	EQUTAB[8] = "SP";
	EQUTAB[9] = "PSW";
	EQUTAB[10]= "AND";
	EQUTAB[11]= "OR";
	EQUTAB[12]= "MOD";
	EQUTAB[13]= "NOT";
	EQUTAB[14]= "XOR";
	EQUTAB[15]= "SHL";
	EQUTAB[16]= "SHR";
	EQUCOUNT = 14;
}


INT ISIDCHR(C)	/* RETURN TRUE IF C IS LEGAL CHARACTER IN IDENTIFIER */
CHAR C;
{	
	RETURN ISALPHA(C) || C == '$' || ISDIGIT(C) || C == '.';
}


INT ISIDSTRT(C)	/* RETURN TRUE IF C IS LEGAL AS FIRST CHAR OF IDENFITIER */
CHAR C;
{
	RETURN ISALPHA(C);
}


INT STREQ(S1, S2)	/* RETURN TRUE IF THE TWO STRINGS ARE EQUAL */
CHAR *S1, *S2;
{
	IF (*S1 != *S2) RETURN 0;	/* SPECIAL CASE FOR SPEED */
	WHILE (*S1) IF (*S1++ != *S2++) RETURN 0;
	RETURN (*S2) ? 0 : 1;
}


SKIP_WSP(STRPTR)	/* SKIP WHITE SPACE AT *STRPTR AND MODIFY THE PTR */
CHAR **STRPTR;
{
	WHILE (ISSPACE(**STRPTR)) (*STRPTR)++;
}


STRCPY2(S1,S2)	/* COPY S2 TO S1, CONVERTING TO UPPER CASE AS WE GO */
CHAR *S1, *S2;
{
	WHILE (*S2)
	     *S1++ = TOUPPER(*S2++);
	*S1 = '\0';
}


/*
	GENERAL-PURPOSE STRING-REPLACEMENT FUNCTION:
		'STRING'	IS POINTER TO ENTIRE STRING,
		'INSSTR'	IS POINTER TO STRING TO BE INSERTED,
		'POS'		IS THE POSITION IN 'STRING' WHERE 'INSSTR'
				IS TO BE INSERTED
		'LENOLD'	IS THE LENGTH OF THE SUBSTRING IN 'STRING'
				THAT IS BEING REPLACED.
*/

REPLSTR(STRING, INSSTR, POS, LENOLD)
CHAR *STRING, *INSSTR;
{
	INT LENGTH, I, J, K, X;

	LENGTH = STRLEN(STRING);
	X = STRLEN(INSSTR);
	K = X - LENOLD;
	I = STRING + POS + LENOLD;
	IF (K) MOVMEM(I, I+K, LENGTH - (POS + LENOLD) + 1);
	FOR (I = 0, J = POS; I < X; I++, J++)
		STRING[J] = INSSTR[I];
}


ERROR(MSG,ARG1,ARG2)
CHAR *MSG;
{
	PRINTF("\N\7%S: %D: ",CFILNAM,LINO);
	PRINTF(MSG,ARG1,ARG2);
	ERRF = 1;
}


ABORT(MSG,ARG1,ARG2)
CHAR *MSG;
{
	ERROR(MSG,ARG1,ARG2);
	PUTCHAR('\N');
	IF (CBUFP == INCBUF) FCLOSE(INCBUF);
	FCLOSE(FBUF);
	IF (*SUBFILE)
		UNLINK(SUBFILE);
	EXIT(-1);
}


SBRK2(N)	/* ALLOCATE STORAGE AND CHECK FOR OUT OF SPACE CONDITION */
{
	INT I;
	IF ((I = SBRK(N)) == ERROR)
		ABORT("OUT OF STORAGE ALLOCATION SPACE\N");
	RETURN I;
}

BUMPTXTP(STR)	/* BUMP TXTBUFP BY SIZE OF GIVEN STRING + 1 */
CHAR *STR;
{
	TXTBUFP += STRLEN(STR) + 1;
	IF (TXTBUFP >= TXTBUF + (TXTBUFSIZE - 8))
	 ABORT("OUT OF TEXT SPACE. INCREASE TXTBUFSIZE AND RECOMPILE CASM");
}


INT NOREL(ID)	/* RETURN TRUE IF IDENTIFIER IS EXEMPT FROM RELOCATETION */
CHAR *ID;
{
	IF (ISEQU(ID)) RETURN 1;
	RETURN 0;
}


INT ISEQU(STR)	/* RETURN TRUE IF GIVEN STRING IS IN THE EQU TABLE */
CHAR *STR;
{
	INT I;
	FOR (I = 0; I < EQUCOUNT; I++)
		IF (STREQ(STR,EQUTAB[I]))
			RETURN 1;
	RETURN 0;
}


CHAR *ISEF(STR)	/* RETURN NFLIST ENTRY IF GIVEN STRING IS AN EXTERNAL */
CHAR *STR;	/* FUNCTION NAME */
{
	INT I;
	FOR (I = 0; I < NFCOUNT; I++)
		IF (STREQ(STR,NFLIST[I]))
			RETURN NFLIST[I];
	RETURN 0;
}

INT HASDOT(STR)	/* RETURN TRUE IF GIVEN STRING HAS A DOT IN IT */
CHAR *STR;
{
	WHILE (*STR)
		IF (*STR++ == '.')
			RETURN TRUE;
	RETURN FALSE;
}

FPUTS2(ARG1,ARG2)
{
	IF (FPUTS(ARG1,ARG2) == ERROR)
		ABORT("OUT OF DISK SPACE FOR OUTPUT FILE.");
}

FPRINTF2(ARG1,ARG2,ARG3,ARG4,ARG5)
{
	IF (FPRINTF(ARG1,ARG2,ARG3,ARG4,ARG5) == ERROR)
		ABORT("OUT OF DISK SPACE FOR OUTPUT FILE.");
}
