/*
	CP.C
	WRITTEN BY LEOR ZOLMAN, 3/82
	MODIFIED FOR PC-1715 BY B.A.   SEP 1986

	CP COPIES FILES TO AND FROM DIFFERENT DRIVES AND USER AREAS ON
	A CP/M FILESYSTEM.

	USAGES:

	A>CP <FILENAME> <NEWNAME> [OPTIONS] <CR>
	A>CP <LIST_OF_FILES> { <USER#/> OR <D:> OR <.> } [OPTIONS] <CR>

	OPTION:		-V	VERIFY BY READING AFTER WRITING

	WILDCARD EXPANSION IS PERFORMED ONLY ON FILES PRESENT IN THE
	CURRENTLY LOGGED DRIVE AND USER AREA (THE "CURRENT DIRECTORY").

	ANY FILENAME MAY CONTAIN AN OPTIONAL DISK DESIGNATOR PREFIX OF
	THE FORM <D:> AND/OR A USER AREA NUMBER PREFIX OF THE FORM <N/>
	PRECEDING THE ACTUAL NAME. IF BOTH MUST BE GIVEN, THE USER AREA
	PREFIX MUST COME FIRST.

	WHEN A SINGLE FILE IS TO BE COPIED, SIMPLY GIVE THE OLD AND NEW
	FILENAMES.

	WHEN MULTIPLE FILES ARE TO BE COPIED, THE DESTINATION PARAMETER
	MUST CONSIST OF EITHER A DISK DESIGNATOR AND/OR USER AREA PATHNAME,
	OR THE SINGLE CHARACTER "." WHEN THE DESTINATION IS THE CURRENT
	DRIVE AND USER AREA. SUCH A DESTINATION DESIGNATION IS ALSO ALLOWED
	WHEN ONLY A SINGLE FILE IS BEING COPIED.

	FOR EXAMPLE, TO COPY *.COM FROM THE CURRENT DIRECTORY INTO USER
	AREA 0 OF DRIVE C:

		CP  *.COM  0/C:

	TO COPY FOO.C FROM USER AREA 4 INTO THE CURRENT USER AREA:

		CP 4/FOO.C .

	COMPILATION AND LINKAGE:
		CC CP.C
		CLINK CP WILDEXP -N
*/

#INCLUDE <STDIO.H>

MAIN(ARGC,ARGV)
CHAR **ARGV;
{
	INT I,J,K,C,LOOP;
	INT FD1,FD2;
	INT BUFSECTS;
	UNSIGNED BUFSIZE;
	UNSIGNED COREBUF;
	CHAR CUR_DISK;			/* CURRENTLY LOGGED-IN DISK */
	CHAR DESTNAME[30];
	CHAR *LASTARG;
	CHAR VERIFY, *VBUF;		/* TRUE WHEN VERIFYING WRITES */

	VERIFY = FALSE;
	WILDEXP(&ARGC,&ARGV);

	CUR_DISK = 'A' + BDOS(25);

	FOR (I = ARGC - 1; I > 2; --I)
		IF (*ARGV[I] == '-')
		{
			SWITCH(ARGV[I][1])
			{
				CASE 'V':
					VERIFY = TRUE;
					BREAK;
			}
			ARGC--;
		}
		ELSE
			BREAK;

	LASTARG = ARGV[ARGC - 1];

	IF (ARGC < 3 || (ARGC > 3 &&
	    !((C = LASTARG[STRLEN(LASTARG)-1]) == '/' || C == ':' ||
					!STRCMP(LASTARG,"."))))
	{
		PRINTF("USAGES: CP [U/][D:]FILENAME [U/][D:]NEWNAME [-V]\N");
		PRINTF("        CP <LIST_OF_FILES>  ");
			PRINTF("{ <U/> OR <D:> OR <U/D:> OR <.>} [-V]\N");
		EXIT();
	}

	IF (VERIFY)
		VBUF =  SBRK(SECSIZ);

	COREBUF = SBRK(SECSIZ);
	FOR (BUFSIZE = SECSIZ; SBRK(SECSIZ) != ERROR; BUFSIZE += SECSIZ)
		;
	BUFSECTS = BUFSIZE / SECSIZ;


   FOR (LOOP = 1; LOOP < ARGC-1; LOOP++)
   {
	IF ((FD1 = OPEN(ARGV[LOOP],0)) == ERROR) {
		PRINTF("\NCAN'T OPEN %S\N",ARGV[LOOP]);
		CONTINUE;	/* GO ON TO NEXT ONE ANYWAY. */
	}

	STRCPY(DESTNAME,LASTARG);	/* CREATE OUTPUT FILENAME */

	IF ( (C = DESTNAME[STRLEN(DESTNAME) - 1])=='/' || C == ':' ||
		  				!STRCMP(DESTNAME,"."))
	{
		IF (!STRCMP(DESTNAME,"."))
			*DESTNAME = '\0';
		FOR (I = STRLEN(ARGV[LOOP]) - 1; I >= 0; I--)
			IF (ARGV[LOOP][I] == '/' || ARGV[LOOP][I] == ':')
				BREAK;				
		STRCAT(DESTNAME,&ARGV[LOOP][I+1]);
	}

	IF ((FD2 = CREAT(DESTNAME)) == ERROR) {
	  	PRINTF("\NCAN'T CREATE %S\N",DESTNAME);
		PRINTF("ASSUMING OUT OF DIRECTORY SPACE AND ABORTING.\N");
		EXIT();	
	}

	IF (LOOP != 1)
		PUTCHAR('\N');

	PRINTF("\T COPYING %S TO %S...",ARGV[LOOP],DESTNAME);

	WHILE (1)
	{
		IF (KBHIT()) GETCHAR();
		IF (!(I = READ(FD1,COREBUF,BUFSECTS))) BREAK;
		IF (I == ERROR)
		{
		    PRINTF("\NREAD ERROR: TELL(FD1) = %D, \"%S\"\N",
						TELL(FD1), ERRMSG(ERRNO()));
		    BREAK;
		}

		IF (KBHIT()) GETCHAR();
		IF (WRITE(FD2,COREBUF,I) != I) {
			PRINTF("\NWRITE ERROR: %S\N",ERRMSG(ERRNO()));
			EXIT();
		}
		IF (VERIFY)
		{	
			PUTS("[V] ");
			SEEK(FD2, -I, 1);
			FOR (J = 0, K = COREBUF; J < I; J++, K += SECSIZ)
			{
				IF (READ(FD2, VBUF, 1) != 1)
					PRINTF("\NVERIFY READ ERROR ON %S\N",
						DESTNAME);
				IF (MEMCMP(VBUF, K, SECSIZ))
					CONTINUE;
				ELSE
					PRINTF("\NVERIFY ERROR ON %S\N",
						DESTNAME);
			}
		}
	}

	IF (CLOSE(FD2) == ERROR) {
		PRINTF("\NCAN'T CLOSE THE OUTPUT FILE.\N");;
		EXIT();
	}
	FABORT(FD1);
   }
}

/*
	RETURN TRUE IF THE STRING ARG IS A FILENAME PREFIXED BY "NN/",
	WHERE "NN" IS A USER NUMBER:
*/

INT HASUNO(STR)
CHAR *STR;
{
	CHAR C;
	INT SUM;

	SUM = 0;

	IF (!ISDIGIT(*STR)) RETURN FALSE;

	WHILE (ISDIGIT(C = *STR++))
		SUM = SUM * 10 + C - '0';
	RETURN (C == '/') ? (SUM >= 0 && SUM < 32) : FALSE;
}
