

/*
	FLOATING POINT PACKAGE SUPPORT ROUTINES

	NOTE THE "FP" LIBRARY FUNCTION, AVAILABLE IN DEFF2.CRL,
	IS USED EXTENSIVELY BY ALL THE FLOATING POINT NUMBER
	CRUNCHING FUNCTIONS.

	-------------------------------------------------------------
	USAGE: AFTER COMPILING YOUR PROGRAM, LINK WITH THIS LIBRARY
	BY TYPING:

	A>CLINK <YOUR PROGRAM FILES> -F FLOAT <CR>
	-------------------------------------------------------------

	NEW FEATURE: A SPECIAL "PRINTF" FUNCTION HAS BEEN INCLUDED
		     IN THIS SOURCE FILE FOR USE WITH FLOATING POINT
		     OPERANDS, IN ADDITION TO THE NORMAL TYPES. THE
		     PRINTF PRESENTED HERE WILL TAKE PRECEDENCE OVER
		     THE DEFF.CRL VERSION WHEN "FLOAT" IS SPECIFIED
		     ON THE CLINK COMMAND LINE AT LINKAGE TIME.
		     NOTE THAT THE "FP" FUNCTION, NEEDED BY MOST OF
		     THE FUNCTIONS IN THIS FILE, RESIDES IN DEFF2.CRL
		     AND WILL BE AUTOMATICALLY COLLECTED BY CLINK.

	ALL FUNCTIONS HERE WRITTEN BY BOB MATHIAS, EXCEPT PRINTF AND
	_SPR (WRITTEN BY LEOR ZOLMAN.)
*/

#INCLUDE <BDSCIO.H>

#DEFINE NORM_CODE	0
#DEFINE ADD_CODE	1
#DEFINE SUB_CODE	2
#DEFINE MULT_CODE	3
#DEFINE DIV_CODE	4
#DEFINE FTOA_CODE	5

FPCOMP(OP1,OP2)
	CHAR *OP1,*OP2;
{
	CHAR WORK[5];
	FPSUB(WORK,OP1,OP2);
	IF (WORK[3] > 127) RETURN (-1);
	IF (WORK[0]+WORK[1]+WORK[2]+WORK[3]) RETURN (1);
	RETURN (0);
}

FPNORM(OP1) CHAR *OP1;
{	FP(NORM_CODE,OP1,OP1);RETURN(OP1);}

FPADD(RESULT,OP1,OP2)
	CHAR *RESULT,*OP1,*OP2;
{	FP(ADD_CODE,RESULT,OP1,OP2);RETURN(RESULT);}

FPSUB(RESULT,OP2,OP1)
	CHAR *RESULT,*OP1,*OP2;
	{FP(SUB_CODE,RESULT,OP1,OP2);RETURN(RESULT);}

FPMULT(RESULT,OP1,OP2)
	CHAR *RESULT,*OP1,*OP2;
{	FP(MULT_CODE,RESULT,OP1,OP2);RETURN(RESULT);}

FPDIV(RESULT,OP1,OP2)
	CHAR *RESULT,*OP1,*OP2;
{	FP(DIV_CODE,RESULT,OP1,OP2);RETURN(RESULT);}

ATOF(FPNO,S)
	CHAR FPNO[5],*S;
{
	CHAR *FPNORM(),WORK[5],ZERO[5],FP_10[5];
	INT SIGN_BOOLEAN,POWER;

	INITB(FP_10,"0,0,0,80,4");
	SETMEM(FPNO,5,0);
	SIGN_BOOLEAN=POWER=0;

	WHILE (*S==' ' || *S=='\T') ++S;
	IF (*S=='-'){SIGN_BOOLEAN=1;++S;}
	FOR (;ISDIGIT(*S);++S){
		FPMULT(FPNO,FPNO,FP_10);
		WORK[0]=*S-'0';
		WORK[1]=WORK[2]=WORK[3]=0;WORK[4]=31;
		FPADD(FPNO,FPNO,FPNORM(WORK));
	}
	IF (*S=='.'){
		++S;
		FOR (;ISDIGIT(*S);--POWER,++S){
			FPMULT(FPNO,FPNO,FP_10);
			WORK[0]=*S-'0';
			WORK[1]=WORK[2]=WORK[3]=0;WORK[4]=31;
			FPADD(FPNO,FPNO,FPNORM(WORK));
		}
	}
	IF (TOUPPER(*S) == 'E') {++S; POWER += ATOI(S); }
	IF (POWER>0)
		FOR (;POWER!=0;--POWER) FPMULT(FPNO,FPNO,FP_10);
	ELSE
	IF (POWER<0)
		FOR (;POWER!=0;++POWER) FPDIV(FPNO,FPNO,FP_10);
	IF (SIGN_BOOLEAN){
		SETMEM(ZERO,5,0);
		FPSUB(FPNO,ZERO,FPNO);
	}
	RETURN(FPNO);
}
FTOA(RESULT,OP1)
	CHAR *RESULT,*OP1;
{	FP(FTOA_CODE,RESULT,OP1);RETURN(RESULT);}

ITOF(OP1,N)
CHAR *OP1;
INT N;
{
	CHAR TEMP[20];
	RETURN ATOF(OP1, ITOA(TEMP,N));
}

ITOA(STR,N)
CHAR *STR;
{
	CHAR *SPTR;
	SPTR = STR;
	IF (N<0) { *SPTR++ = '-'; N = -N; }
	_USPR(&SPTR, N, 10);
	*SPTR = '\0';
	RETURN STR;
}

/*
	THIS IS THE SPECIAL FORMATTING FUNCTION, WHICH SUPPORTS THE
	"E" AND "F" CONVERSIONS AS WELL AS THE NORMAL "D", "S", ETC.
	WHEN USING "E" OR "F" FORMAT, THE CORRESPONDING ARGUMENT IN
	THE ARGUMENT LIST SHOULD BE A POINTER TO ONE OF THE FIVE-BYTE
	STRINGS USED AS FLOATING POINT NUMBERS BY THE FLOATING POINT
	FUNCTIONS. NOTE THAT YOU DON'T NEED TO EVER USE THE "FTOA"
	FUNCTION WHEN USING THIS SPECIAL PRINTF/SPRINTF COMBINATION;
	TO ACHIEVE THE SAME RESULT AS FTOA, A SIMPLE "%E" FORMAT
	CONVERSION WILL DO THE TRICK. "%F" IS USED TO ELIMINATE THE
	SCIENTIFIC NOTATION AND SET THE PRECISION. THE ONLY [KNOWN]
	DIFFERENCE BETWEEN THE "E" AND "F" CONVERSIONS AS USED HERE
	AND THE ONES DESCRIBED IN THE KERNIGHAN & RITCHIE BOOK IS THAT
	ROUNDING DOES NOT TAKE PLACE IN THIS VERSION...E.G., PRINTING
	A FLOATING POINT NUMBER WHICH HAPPENS TO EQUAL EXACTLY 3.999
	USING A "%5.2F" FORMAT CONVERSION WILL PRODUCE " 3.99" INSTEAD
	OF " 4.00".
*/

PRINTF(FORMAT)
CHAR *FORMAT;
{
	INT PUTCHAR();
	_SPR(&FORMAT, &PUTCHAR);	/* USE "_SPR" TO FORM THE OUTPUT */
}


_SPR(FMT,PUTCF,ARG1)
INT (*PUTCF)();
CHAR **FMT;
{
	CHAR _USPR(), C, BASE, *SPTR, *FORMAT;
	CHAR WBUF[MAXLINE], *WPTR, PF, LJFLAG, ZFFLAG;
	INT WIDTH, PRECISION, EXP, *ARGS;

	FORMAT = *FMT++;	/* FMT FIRST POINTS TO THE FORMAT STRING */
	ARGS = FMT;		/* NOW FMT POINTS TO THE FIRST ARG VALUE */
	WHILE (C = *FORMAT++)
	  IF (C == '%') {
	    WPTR = WBUF;
	    PRECISION = 6;
	    LJFLAG = PF = ZFFLAG = 0;

	    IF (*FORMAT == '-') {
		    FORMAT++;
		    LJFLAG++;
	     }

	    IF (*FORMAT == '0') ZFFLAG++;	/* TEST FOR ZERO FILL */

	    WIDTH = ISDIGIT(*FORMAT) ? _GV2(&FORMAT) : 0;

	    IF ((C = *FORMAT++) == '.') {
		    PRECISION = _GV2(&FORMAT);
		    PF++;
		    C = *FORMAT++;
	     }

	    SWITCH(TOUPPER(C)) {
		CASE 'E':  IF (PRECISION>7) PRECISION = 7;
			   FTOA(WBUF,*ARGS++);
			   STRCPY(WBUF+PRECISION+3, WBUF+10);
			   WIDTH -= STRLEN(WBUF);
			   GOTO PAD2;

		CASE 'F':  FTOA(&WBUF[60],*ARGS++);
			   SPTR = &WBUF[60];
			   WHILE ( *SPTR++ != 'E')
				;
			   EXP = ATOI(SPTR);
			   SPTR = &WBUF[60];
			   IF (*SPTR == ' ') SPTR++;
			   IF (*SPTR == '-') {
				*WPTR++ = '-';
				SPTR++;
				WIDTH--;
			    }
			   SPTR += 2;

			   IF (EXP < 1) {
				*WPTR++ = '0';
				WIDTH--;
			    }

			   PF = 7;
			   WHILE (EXP > 0 && PF) {
				*WPTR++ = *SPTR++;
				PF--;
				EXP--;
				WIDTH--;
			    }

			   WHILE (EXP > 0) {
				*WPTR++ = '0';
				EXP--;
				WIDTH--;
			    }

			   *WPTR++ = '.';
			   WIDTH--;

			   WHILE (EXP < 0 && PRECISION) {
				*WPTR++ = '0';
				EXP++;
				PRECISION--;
				WIDTH--;
			    }

			   WHILE (PRECISION && PF) {
				*WPTR++ = *SPTR++;
				PF--;
				PRECISION--;
				WIDTH--;
			    }

			   WHILE (PRECISION>0) {
				*WPTR++ = '0';
				PRECISION--;
				WIDTH--;
			    }

			   GOTO PAD;


		CASE 'D':  IF (*ARGS < 0) {
				*WPTR++ = '-';
				*ARGS = -*ARGS;
				WIDTH--;
			    }
		CASE 'U':  BASE = 10; GOTO VAL;

		CASE 'X':  BASE = 16; GOTO VAL;

		CASE 'O':  BASE = 8;

		     VAL:  WIDTH -= _USPR(&WPTR,*ARGS++,BASE);
			   GOTO PAD;

		CASE 'C':  *WPTR++ = *ARGS++;
			   WIDTH--;
			   GOTO PAD;

		CASE 'S':  IF (!PF) PRECISION = 200;
			   SPTR = *ARGS++;
			   WHILE (*SPTR && PRECISION) {
				*WPTR++ = *SPTR++;
				PRECISION--;
				WIDTH--;
			    }

		     PAD:  *WPTR = '\0';
		     PAD2: WPTR = WBUF;
			   IF (!LJFLAG)
			    WHILE (WIDTH-- > 0)
				IF ((*PUTCF)(ZFFLAG ? '0' : ' ',ARG1) == ERROR)
					RETURN ERROR;

			    WHILE (*WPTR)
				IF ((*PUTCF)(*WPTR++, ARG1) == ERROR)
					RETURN ERROR;

			    IF (LJFLAG)
			     WHILE (WIDTH-- > 0)
				IF ((*PUTCF)(' ', ARG1) == ERROR)
					RETURN ERROR;
			   BREAK;

		 DEFAULT:  IF ((*PUTCF)(C, ARG1) == ERROR)
				RETURN ERROR;
	     }
	  }
	  ELSE IF ((*PUTCF)(C, ARG1) == ERROR)
		RETURN ERROR;
}

