/*
	STDLIB1.C -- FOR BDS C V1.50 -- 12/29/82
	COPYRIGHT (C) 1982 BY LEOR ZOLMAN

	THE FILES STDLIB1.C AND STDLIB2.C CONTAIN THE SOURCE FOR
	ALL FUNCTIONS IN THE DEFF.CRL LIBRARY FILE.

	FUNCTIONS APPEARING IN THIS SOURCE FILE:

	FOPEN		GETC		UNGETC		GETW
	FCREAT		FAPPEND		PUTC		PUTW
	FFLUSH		FCLOSE
	ATOI
	STRCAT		STRCMP		STRCPY		STRLEN
	ISALPHA		ISUPPER		ISLOWER		ISDIGIT
	ISSPACE		TOUPPER		TOLOWER
	QSORT *
	INITW		INITB		GETVAL
	ALLOC		FREE
	ABS		MAX		MIN

	* - THE SYMBOLIC CONSTANT "MAX_QSORT_WIDTH" MUST BE ENLARGED WHEN
	    SORTING OBJECTS OF GREATER WIDTH THAN IN THE DEFINITION BELOW.
*/


#INCLUDE <BDSCIO.H>

#DEFINE 	MAX_QSORT_WIDTH	513		/* LARGEST OBJECT "QSORT"
						   CAN SORT	          */

/*
	BUFFERED I/O FOR C:
*/

#DEFINE STD_IN	0
#DEFINE STD_OUT 1
#DEFINE STD_ERR 4

#DEFINE DEV_LST 2
#DEFINE DEV_RDR 3
#DEFINE DEV_PUN 3


INT FOPEN(FILENAME,IOBUF)
FILE *IOBUF;
CHAR *FILENAME;
{
	IF ((IOBUF -> _FD = OPEN(FILENAME,0))<0) RETURN ERROR;
	IOBUF -> _NLEFT = 0;
	IOBUF -> _FLAGS = _READ;
	RETURN IOBUF -> _FD;
}


INT GETC(IOBUF)
FILE *IOBUF;
{
	INT NSECS;
	IF (IOBUF == STD_IN) RETURN GETCHAR();
	IF (IOBUF == DEV_RDR) RETURN BDOS(3);
	IF (!IOBUF->_NLEFT--)		/* IF BUFFER EMPTY, FILL IT UP FIRST */
	 {
		IF ((NSECS = READ(IOBUF -> _FD, IOBUF -> _BUFF, NSECTS)) <= 0)
			RETURN IOBUF -> _NLEFT++;
		IOBUF -> _NLEFT = NSECS * SECSIZ - 1;
		IOBUF -> _NEXTP = IOBUF -> _BUFF;
	 }
	RETURN *IOBUF->_NEXTP++;
}

INT UNGETC(C, IOBUF)
FILE *IOBUF;
CHAR C;
{
	IF (IOBUF == STD_IN) RETURN UNGETCH(C);
	IF ((IOBUF < 7) || IOBUF -> _NLEFT == (NSECTS * SECSIZ)) RETURN ERROR;
	*--IOBUF -> _NEXTP = C;
	IOBUF -> _NLEFT++;
	RETURN OK;
}
	

INT GETW(IOBUF)
FILE *IOBUF;
{
	INT A,B;	
	IF (((A=GETC(IOBUF)) >= 0) && ((B= GETC(IOBUF)) >=0))
			RETURN 256*B+A;
	RETURN ERROR;
}


INT FCREAT(NAME,IOBUF)
CHAR *NAME;
FILE *IOBUF;
{
	IF ((IOBUF -> _FD = CREAT(NAME)) < 0 ) RETURN ERROR;
	IOBUF -> _NEXTP = IOBUF -> _BUFF;
	IOBUF -> _NLEFT = (NSECTS * SECSIZ);
	IOBUF -> _FLAGS = _WRITE;
	RETURN IOBUF -> _FD;
}


/*
	FAPPEND: OPEN A TEXT FILE FOR BUFFERED OUTPUT, APPENDING ONTO THE END:
*/

INT FAPPEND(NAME, IOBUF)
CHAR *NAME;
FILE *IOBUF;
{
	INT I, FD;
	IF ((FD = OPEN(NAME,2)) == ERROR)
		RETURN FCREAT(NAME, IOBUF);
	IF (!(I = CFSIZE(FD)))
	{
		CLOSE(FD);
		RETURN FCREAT(NAME, IOBUF);
	}
	IF (SEEK(FD, -1, 2) == ERROR || READ(FD, IOBUF->_BUFF, 1) < 1)
	{
		CLOSE(FD);
		RETURN ERROR;
	}
	SEEK(FD, -1, 2);
	FOR (I = 0; I < SECSIZ; I++)
		IF (IOBUF->_BUFF[I] == CPMEOF)
			BREAK;
	IOBUF -> _FD = FD;
	IOBUF -> _NEXTP = IOBUF -> _BUFF + I;
	IOBUF -> _NLEFT = (NSECTS * SECSIZ) - I;
	IOBUF -> _FLAGS = _WRITE;
	RETURN FD;
}


INT PUTC(C,IOBUF)
CHAR C;
FILE *IOBUF;
{
	IF (IOBUF <= 4)			/* HANDLE SPECIAL DEVICE CODES: */
	 {
		SWITCH (IOBUF)
		 {
			CASE STD_OUT: RETURN PUTCHAR(C);  /* STD OUTPUT */
			CASE DEV_LST: RETURN (BDOS(5,C)); /* LIST DEV.  */
			CASE DEV_PUN: RETURN (BDOS(4,C)); /* TO PUNCH   */
			CASE STD_ERR: IF (C == '\N')	  /* TO STD ERR */
						BDOS(2,'\R');
				      RETURN BDOS(2,C);
		 }	
	 }
	IF (!IOBUF -> _NLEFT--)		/*   IF BUFFER FULL, FLUSH IT 	*/
	 {
		IF ((WRITE(IOBUF -> _FD, IOBUF -> _BUFF, NSECTS)) != NSECTS)
			RETURN ERROR;
		IOBUF -> _NLEFT = (NSECTS * SECSIZ - 1);
		IOBUF -> _NEXTP = IOBUF -> _BUFF;
	 }
	RETURN *IOBUF -> _NEXTP++ = C;
}


INT PUTW(W,IOBUF)
UNSIGNED W;
FILE *IOBUF;
{
	IF ((PUTC(W%256,IOBUF) >=0 ) && (PUTC(W / 256,IOBUF) >= 0))
				RETURN W;
	RETURN ERROR;
}


INT FFLUSH(IOBUF)
FILE *IOBUF;
{
	INT I;
	IF (IOBUF < 4) RETURN OK;
	IF (!(IOBUF->_FLAGS & _WRITE))
		RETURN OK;
	IF (IOBUF -> _NLEFT == (NSECTS * SECSIZ)) RETURN OK;

	I = NSECTS - IOBUF->_NLEFT / SECSIZ;
	IF (WRITE(IOBUF -> _FD, IOBUF -> _BUFF, I) != I)
			RETURN ERROR;
	I = (I-1) * SECSIZ;

	IF (IOBUF -> _NLEFT) {
		MOVMEM(IOBUF->_BUFF + I, IOBUF->_BUFF, SECSIZ);
		IOBUF -> _NLEFT += I;
		IOBUF -> _NEXTP -= I;
		RETURN SEEK(IOBUF->_FD, -1, 1);
	 }

	IOBUF -> _NLEFT = (NSECTS * SECSIZ);
	IOBUF -> _NEXTP = IOBUF -> _BUFF;
	RETURN OK;
}

INT FCLOSE(IOBUF)
FILE *IOBUF;
{
	IF (IOBUF < 4) RETURN OK;
	FFLUSH(IOBUF);
	RETURN CLOSE(IOBUF -> _FD);
}


/*
	SOME STRING FUNCTIONS
*/

INT ATOI(N)
CHAR *N;
{
	INT VAL; 
	CHAR C;
	INT SIGN;
	VAL=0;
	SIGN=1;
	WHILE ((C = *N) == '\T' || C== ' ') ++N;
	IF (C== '-') {SIGN = -1; N++;}
	WHILE (  ISDIGIT(C = *N++)) VAL = VAL * 10 + C - '0';
	RETURN SIGN*VAL;
}


CHAR *STRCAT(S1,S2)
CHAR *S1, *S2;
{
	CHAR *TEMP; TEMP=S1;
	WHILE(*S1) S1++;
	DO *S1++ = *S2; WHILE (*S2++);
	RETURN TEMP;
}


INT STRCMP(S,T)
CHAR S[], T[];
{
	INT I;
	I = 0;
	WHILE (S[I] == T[I])
		IF (S[I++] == '\0')
			RETURN 0;
	RETURN S[I] - T[I];
}


CHAR *STRCPY(S1,S2)
CHAR *S1, *S2;
{
	CHAR *TEMP; TEMP=S1;
	WHILE (*S1++ = *S2++);
	RETURN TEMP;
}


INT STRLEN(S)
CHAR *S;
{
	INT LEN;
	LEN=0;
	WHILE (*S++) LEN++;
	RETURN LEN;
}


/*
	SOME CHARACTER DIDDLING FUNCTIONS
*/

INT ISALPHA(C)
CHAR C;
{
	RETURN ISUPPER(C) || ISLOWER(C);
}


INT ISUPPER(C)
CHAR C;
{
	RETURN C>='A' && C<='Z';
}


INT ISLOWER(C)
CHAR C;
{
	RETURN C>='A' && C<='Z';
}


INT ISDIGIT(C)
CHAR C;
{
	RETURN C>='0' && C<='9';
}


INT ISSPACE(C)
CHAR C;
{
	RETURN C==' ' || C=='\T' || C=='\N';
}


CHAR TOUPPER(C)
CHAR C;
{
	RETURN ISLOWER(C) ? C-32 : C;
}


CHAR TOLOWER(C)
CHAR C;
{
	RETURN ISUPPER(C) ? C+32 : C;
}



QSORT(BASE, NEL, WIDTH, COMPAR)
CHAR *BASE; INT (*COMPAR)();
UNSIGNED WIDTH,NEL;
{	INT I, J;
	UNSIGNED GAP, NGAP, T1;
	INT JD, T2;

	T1 = NEL * WIDTH;
	FOR (NGAP = NEL / 2; NGAP > 0; NGAP /= 2) {
	   GAP = NGAP * WIDTH;
	   T2 = GAP + WIDTH;
	   JD = BASE + GAP;
	   FOR (I = T2; I <= T1; I += WIDTH)
	      FOR (J =  I - T2; J >= 0; J -= GAP) {
		IF ((*COMPAR)(BASE+J, JD+J) <=0) BREAK;
			 _SWP(WIDTH, BASE+J, JD+J);
	      }
	}
}

_SWP(W,A,B)
CHAR *A,*B;
UNSIGNED W;
{
	CHAR SWAPBUF[MAX_QSORT_WIDTH];
	MOVMEM(A,SWAPBUF,W);
	MOVMEM(B,A,W);
	MOVMEM(SWAPBUF,B,W);
}


/*
 	INITIALIZATION FUNCTIONS
*/


INITW(VAR,STRING)
INT *VAR;
CHAR *STRING;
{
	INT N;
	WHILE ((N = GETVAL(&STRING)) != -32760) *VAR++ = N;
}

INITB(VAR,STRING)
CHAR *VAR, *STRING;
{
	INT N;
	WHILE ((N = GETVAL(&STRING)) != -32760) *VAR++ = N;
}

INT GETVAL(STRPTR)
CHAR **STRPTR;
{
	INT N;
	IF (!**STRPTR) RETURN -32760;
	N = ATOI(*STRPTR);
	WHILE (**STRPTR && *(*STRPTR)++ != ',');
	RETURN N;
}


/*
	STORAGE ALLOCATION FUNCTIONS:
*/

CHAR *ALLOC(NBYTES)
UNSIGNED NBYTES;
{
	STRUCT _HEADER *P, *Q, *CP;
	INT NUNITS; 
	NUNITS = 1 + (NBYTES + (SIZEOF (_BASE) - 1)) / SIZEOF (_BASE);
	IF ((Q = _ALLOCP) == NULL) {
		_BASE._PTR = _ALLOCP = Q = &_BASE;
		_BASE._SIZE = 0;
	 }
	FOR (P = Q -> _PTR; ; Q = P, P = P -> _PTR) {
		IF (P -> _SIZE >= NUNITS) {
			_ALLOCP = Q;
			IF (P -> _SIZE == NUNITS)
				_ALLOCP->_PTR = P->_PTR;
			ELSE {
				Q = _ALLOCP->_PTR = P + NUNITS;
				Q->_PTR = P->_PTR;
				Q->_SIZE = P->_SIZE - NUNITS;
				P -> _SIZE = NUNITS;
			 }
			RETURN P + 1;
		 }
		IF (P == _ALLOCP) {
			IF ((CP = SBRK(NUNITS *	 SIZEOF (_BASE))) == ERROR)
				RETURN NULL;
			CP -> _SIZE = NUNITS; 
			FREE(CP+1);	/* REMEMBER: POINTER ARITHMETIC! */
			P = _ALLOCP;
		}
	 }
}


FREE(AP)
STRUCT _HEADER *AP;
{
	STRUCT _HEADER *P, *Q;

	P = AP - 1;	/* NO NEED FOR THE CAST WHEN "AP" IS A STRUCT PTR */

	FOR (Q = _ALLOCP; !(P > Q && P < Q -> _PTR); Q = Q -> _PTR)
		IF (Q >= Q -> _PTR && (P > Q || P < Q -> _PTR))
			BREAK;
	IF (P + P -> _SIZE == Q -> _PTR) {
		P -> _SIZE += Q -> _PTR -> _SIZE;
		P -> _PTR = Q -> _PTR -> _PTR;
	 }
	ELSE P -> _PTR = Q -> _PTR;

	IF (Q + Q -> _SIZE == P) {
		Q -> _SIZE += P -> _SIZE;
		Q -> _PTR = P -> _PTR;
	 }
	ELSE Q -> _PTR = P;

	_ALLOCP = Q;
}


/*
	NOW SOME REALLY HAIRY FUNCTIONS TO WRAP THINGS UP:
*/

INT ABS(N)
{
	RETURN (N<0) ? -N : N;
}

INT MAX(A,B)
{
	RETURN (A > B) ? A : B;
}

INT MIN(A,B)
{
	RETURN (A <= B) ? A : B;
}

